@@ -22,7 +22,6 @@ class Entry:
     # cname            string     C name of entity
     # type             PyrexType  Type of entity
     # doc              string     Doc string
-    # #borrowed         bool       Is a borrowed reference
     # init             string     Initial value
     # visibility       'private' or 'public' or 'extern'
     # is_builtin       boolean    Is a Python builtin name
@@ -62,6 +61,7 @@ class Entry:
     # interned_cname   string     C name of interned name string
     # pystring_cname   string     C name of Python version of string literal
     # is_interned      boolean    For string const entries, value is interned
+    # used             boolean
 
     borrowed = 0
     init = ""
@@ -92,6 +92,7 @@ class Entry:
     interned_cname = None
     pystring_cname = None
     is_interned = 0
+    used = 0
     
     def __init__(self, name, cname, type, pos = None, init = None):
         self.name = name
@@ -171,7 +172,6 @@ class Scope:
         return self.global_scope().intern(name)
     
     def qualifying_scope(self):
-        #return self.outer_scope
         return self.parent_scope
     
     def mangle(self, prefix, name = None):
@@ -208,9 +208,6 @@ class Scope:
     def qualify_name(self, name):
         return "%s.%s" % (self.qualified_name, name)
     
-    #def undeclare(self, name):
-    #	del self.entries[name]
-    
     def declare_const(self, name, type, value, pos, cname = None):
         # Add an entry for a named constant.
         if not cname:
@@ -356,6 +353,7 @@ class Scope:
         # Add an entry for a string constant.
         cname = self.new_const_cname()
         entry = Entry("", cname, c_char_array_type, init = value)
+        entry.used = 1
         self.const_entries.append(entry)
         return entry
     
@@ -400,6 +398,7 @@ class Scope:
         self.temp_counter = n + 1
         cname = "%s%d" % (Naming.pyrex_prefix, n)
         entry = Entry("", cname, type)
+        entry.used = 1
         if type.is_pyobject:
             entry.init = "0"
         self.cname_to_entry[entry.cname] = entry
@@ -420,17 +419,23 @@ class Scope:
                 % cname)
         self.free_temp_entries.append(entry)
     
-    def recycle_pending_temps(self):
-        # Obsolete
-        pass
+    def temps_in_use(self):
+        # Return a new list of temp entries currently in use.
+        return [entry for entry in self.temp_entries
+            if entry not in self.free_temp_entries]
+    
+    #def recycle_pending_temps(self):
+    #	# Obsolete
+    #	pass
 
     def use_utility_code(self, new_code):
         self.global_scope().use_utility_code(new_code)
     
     def generate_library_function_declarations(self, code):
         # Generate extern decls for C library funcs used.
-        if self.pow_function_used:
-            code.putln("extern double pow(double, double);")
+        #if self.pow_function_used:
+        #	code.putln("%s double pow(double, double);" % Naming.extern_c_macro)
+        pass
         
     def defines_any(self, names):
         # Test whether any of the given names are
@@ -475,6 +480,7 @@ class ModuleScope(Scope):
     # intern_map           {string : string}  Mapping from Python names to interned strs
     # interned_names       [string]           Interned names pending generation of declarations
     # all_pystring_entries [Entry]            Python string consts from all scopes
+    # types_imported       {PyrexType : 1}    Set of types for which import code generated
 
     def __init__(self, name, parent_module, context):
         self.parent_module = parent_module
@@ -499,6 +505,7 @@ class ModuleScope(Scope):
         self.intern_map = {}
         self.interned_names = []
         self.all_pystring_entries = []
+        self.types_imported = {}
     
     def qualifying_scope(self):
         return self.parent_module
@@ -564,6 +571,8 @@ class ModuleScope(Scope):
         # None if previously declared as something else.
         entry = self.lookup_here(name)
         if entry:
+            if entry.is_pyglobal and entry.as_module is scope:
+                return entry # Already declared as the same module
             if not (entry.is_pyglobal and not entry.as_module):
                 error(pos, "'%s' redeclared" % name)
                 return None
@@ -885,8 +894,8 @@ class PyClassScope(ClassScope):
     def release_temp(self, cname):
         self.outer_scope.release_temp(cname)
 
-    def recycle_pending_temps(self):
-        self.outer_scope.recycle_pending_temps()
+    #def recycle_pending_temps(self):
+    #	self.outer_scope.recycle_pending_temps()
 
     def add_default_value(self, type):
         return self.outer_scope.add_default_value(type)
@@ -955,6 +964,8 @@ class CClassScope(ClassScope):
         if visibility in ('public', 'readonly'):
             if type.pymemberdef_typecode:
                 self.public_attr_entries.append(entry)
+                if name == "__weakref__":
+                    error(pos, "Special attribute __weakref__ cannot be exposed to Python")
             else:
                 error(pos,
                     "C attribute of type '%s' cannot be accessed from Python" % type)
@@ -977,6 +988,8 @@ class CClassScope(ClassScope):
             
     def declare_cfunction(self, name, type, pos,
             cname = None, visibility = 'private', defining = 0):
+        if get_special_method_signature(name):
+            error(pos, "Special methods must be declared with 'def', not 'cdef'")
         args = type.args
         if not args:
             error(pos, "C method has no self argument")
